"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createReplicationMultiConcurrencyController = void 0;
const error_1 = require("../../common/error");
const create_discriminating_mutex_concurrency_controller_1 = require("./create-discriminating-mutex-concurrency-controller");
const create_full_concurrency_controller_1 = require("./create-full-concurrency-controller");
const create_mutex_concurrency_controller_1 = require("./create-mutex-concurrency-controller");
const create_semaphore_concurrency_controller_1 = require("./create-semaphore-concurrency-controller");
/**
 * Use different mutex controllers depending on the desired concurrency level
 * for different messages. If a discriminating mutex should be used, the
 * discriminator function must be supplied as well. If a semaphore concurrency
 * controller should be used, the `maxSemaphoreParallelism` should be set (it
 * defaults to 5);
 * @param getConcurrencyType Implements the logic which concurrency controller should be used e.g. based on the aggregate and message type
 * @param discriminator The discriminator to find or create a mutex for when using the discriminating mutex.
 * @returns The controller to acquire and release the mutex for a specific discriminator
 */
const createReplicationMultiConcurrencyController = (getConcurrencyType, settings) => {
    var _a;
    const fullConcurrencyController = (0, create_full_concurrency_controller_1.createReplicationFullConcurrencyController)();
    const mutexController = (0, create_mutex_concurrency_controller_1.createReplicationMutexConcurrencyController)();
    const discriminatingMutexController = (settings === null || settings === void 0 ? void 0 : settings.discriminator)
        ? (0, create_discriminating_mutex_concurrency_controller_1.createReplicationDiscriminatingMutexConcurrencyController)(settings.discriminator)
        : undefined;
    const semaphore = (0, create_semaphore_concurrency_controller_1.createReplicationSemaphoreConcurrencyController)((_a = settings === null || settings === void 0 ? void 0 : settings.maxSemaphoreParallelism) !== null && _a !== void 0 ? _a : 5);
    return {
        /** Acquire a lock (if any) and return a function to release it. */
        acquire: (message) => {
            switch (getConcurrencyType(message)) {
                case 'full-concurrency':
                    return fullConcurrencyController.acquire(message);
                case 'mutex':
                    return mutexController.acquire(message);
                case 'semaphore':
                    return semaphore.acquire(message);
                case 'discriminating-mutex':
                    if (!discriminatingMutexController) {
                        throw new error_1.MessageError('A discriminating mutex controller was not configured.', 'DISCRIMINATING_MUTEX_NOT_CONFIGURED', message);
                    }
                    return discriminatingMutexController.acquire(message);
            }
        },
        /** Cancel all controllers. */
        cancel: () => {
            fullConcurrencyController.cancel();
            mutexController.cancel();
            discriminatingMutexController === null || discriminatingMutexController === void 0 ? void 0 : discriminatingMutexController.cancel();
            semaphore.cancel();
        },
    };
};
exports.createReplicationMultiConcurrencyController = createReplicationMultiConcurrencyController;
//# sourceMappingURL=create-multi-concurrency-controller.js.map