#! /usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.dbSetupCli = void 0;
const promises_1 = require("node:fs/promises");
const readline_1 = require("readline");
const database_setup_exporter_1 = require("./database-setup-exporter");
/* eslint-disable @typescript-eslint/no-var-requires */
const rli = (0, readline_1.createInterface)({
    input: process.stdin,
    output: process.stdout,
});
const { createPollingScript, createReplicationScript } = database_setup_exporter_1.DatabaseSetupExporter;
/** Async way to ask a question from the CLI */
const input = (prompt) => {
    return new Promise((callbackFn) => {
        rli.question(prompt, (userInput) => {
            callbackFn(userInput === null || userInput === void 0 ? void 0 : userInput.trim());
        });
    });
};
/** Get a value from the command line with allowed values and a default value */
const getValue = (prompt, allowedAnswers, defaultValue) => __awaiter(void 0, void 0, void 0, function* () {
    let answer;
    do {
        const d = defaultValue ? ` (default: ${defaultValue})` : '';
        answer = yield input(`\x1b[32m${prompt}\x1b[0m${d}\n> `);
        if (defaultValue && !answer) {
            answer = defaultValue;
        }
    } while (allowedAnswers ? allowedAnswers.indexOf(answer) < 0 : !answer);
    return answer;
});
/** Get config values for the logical replication listener */
const logicalReplicationValues = (outboxOrInbox) => __awaiter(void 0, void 0, void 0, function* () {
    const replicationSlot = yield getValue(`What should the name of the replication slot for the ${outboxOrInbox} be?`, undefined, `transactional_${outboxOrInbox}_slot`);
    const publication = yield getValue(`What should the name of the replication publication for the ${outboxOrInbox} be?`, undefined, `transactional_${outboxOrInbox}_publication`);
    return { replicationSlot, publication };
});
/** Get config values for the polling listener */
const pollingValues = (outboxOrInbox, schema) => __awaiter(void 0, void 0, void 0, function* () {
    const nextMessagesName = yield getValue(`What should the name of the ${outboxOrInbox} polling function be?`, undefined, `next_${outboxOrInbox}_messages`);
    const nextMessagesSchema = yield getValue(`What should the name of the database schema for the "${nextMessagesName}" function be?`, undefined, schema);
    return { nextMessagesName, nextMessagesSchema };
});
/** Execute the CLI */
const dbSetupCli = () => __awaiter(void 0, void 0, void 0, function* () {
    const listener = yield getValue('What type of listener should be created?\n- Logical Replication (r)\n- Polling (p)', ['r', 'p']);
    const database = yield getValue('What is the name of your database?');
    const schema = yield getValue('What should the name of the database schema be?', undefined, 'messaging');
    const listenerRole = yield getValue('What is the name of your database role for the listener?', undefined, 'messaging_listener');
    const handlerRole = yield getValue('What is the name of your database role for the message handler?', undefined, listenerRole);
    const outboxInboxBoth = yield getValue('What transactional types should be created?\n- outbox (o)\n- inbox (i)\n- both (b)', ['o', 'i', 'b']);
    const buildReplicationSql = (outboxOrInbox, table, skipRoles = false) => __awaiter(void 0, void 0, void 0, function* () {
        const { publication, replicationSlot } = yield logicalReplicationValues(outboxOrInbox);
        const config = {
            database,
            schema,
            table,
            listenerRole,
            handlerRole,
            outboxOrInbox,
            publication,
            replicationSlot,
        };
        return createReplicationScript(config, skipRoles);
    });
    const buildPollingSql = (outboxOrInbox, table, skipRoles = false) => __awaiter(void 0, void 0, void 0, function* () {
        const { nextMessagesName, nextMessagesSchema } = yield pollingValues(outboxOrInbox, schema);
        const config = {
            database,
            schema,
            table,
            listenerRole,
            handlerRole,
            outboxOrInbox,
            nextMessagesName,
            nextMessagesSchema,
        };
        return createPollingScript(config, skipRoles);
    });
    let sqlOutput = '';
    if (outboxInboxBoth === 'o' || outboxInboxBoth === 'b') {
        const outboxOrInbox = 'outbox';
        const table = yield getValue('What name should the outbox table have?', undefined, outboxOrInbox);
        if (listener === 'r') {
            sqlOutput += yield buildReplicationSql(outboxOrInbox, table);
        }
        else {
            sqlOutput += yield buildPollingSql(outboxOrInbox, table);
        }
    }
    if (outboxInboxBoth === 'i' || outboxInboxBoth === 'b') {
        const outboxOrInbox = 'inbox';
        const table = yield getValue('What name should the inbox table have?', undefined, outboxOrInbox);
        if (listener === 'r') {
            sqlOutput += yield buildReplicationSql(outboxOrInbox, table, outboxInboxBoth === 'b');
        }
        else {
            sqlOutput += yield buildPollingSql(outboxOrInbox, table, outboxInboxBoth === 'b');
        }
    }
    const filename = yield getValue('What should the filename for the SQL script be?', undefined, 'transactional.sql');
    yield (0, promises_1.writeFile)(filename, sqlOutput);
    console.log(`File \x1b[92m${filename}\x1b[0m successfully created.`);
    rli.close();
});
exports.dbSetupCli = dbSetupCli;
//# sourceMappingURL=db-setup-cli.js.map